#!/usr/bin/perl
# A Perl text editor.
# Copyright (c) 1998 by unSlider, all rights reserved.
# You may use and distribute this under the terms of the Perl Artistic License.

use Curses;

$acedver=19981209.0;		# our version number
@buffer=();
$cbuf=\@buffer;			# current buffer
%info=();
$cinfo=\%info;			# settings that apply to current buffer

%defbindings=(			# default key bindings
	      "\cW" => sub {$diedie=1;},
	      "\c[" => sub {
		  my $inchar=getch;
		  
		  if($repeated && ($inchar ne $lastmeta)) {
		      $repeated=0;
		  }
		  if(defined($$cinfo{metabindings}{$inchar})) {
		      &{$$cinfo{metabindings}{$inchar}};
		  } else {
		      &errmsg("no such meta binding $inchar");
		  }
		  $lastmeta=$inchar;
	      },
	      "\cf" => \&cursfwd,
	      @{[KEY_RIGHT]} => \&cursfwd,
	      "\cb" => \&cursback,
	      @{[KEY_LEFT]} => \&cursback,
	      "\ca" => \&cursbol,
	      @{[KEY_HOME]} => \&cursbol,
	      "\ce" => \&curseol,
	      @{[KEY_END]} => \&curseol,
	      "\cj" => \&insnewline,
	      "\ch" => \&backspace,
	      @{[KEY_BACKSPACE]} => \&backspace,
	      "\cd" => \&delunder,
	      @{[KEY_DC]} => \&delunder,
	      "\cn" => \&cursdown,
	      @{[KEY_DOWN]} => \&cursdown,
	      "\cp" => \&cursup,
	      @{[KEY_UP]} => \&cursup,
	      "\cv" => \&curspgdn,
	      @{[KEY_NPAGE]} => \&curspgdn,
	      "\cy" => \&curspgup,
	      @{[KEY_PPAGE]} => \&curspgup,
	      "\ck" => \&cutline,
	      "\cu" => sub {&paste($cutbuffer)},
	      "\ci" => \&tab,
	   );

%defmetabindings=(		# default meta-key bindings
		  "z" => sub {&statmsg("meta-z was pressed");},
		  "c" => sub {
		      eval(&getline("> "));
		      if($@) {
			  &errmsg($@);
		      }
		      &movecurs();
		  },
		  "f" => \&find,
		  "s" => \&saveas,
		  "l" => \&load,
		  "\cA" => \&cursbof,
		  "\cE" => \&curseof,
	       );

$info{bindings}=\%defbindings;
$info{metabindings}=\%defmetabindings;
$info{wintop}=0;
$info{statlines}=1;
$info{statfunc}=\&statbar;
# cursy is the offset into the CURRENTLY DISPLAYED WINDOW of the current line
# position is the offset of the first line of the window into the buffer
# cursx is the offset, IN FILE (not display) CHARACTERS, into the BUFFER VIEW
# OF THE CURRENT LINE (ie, left/right panning does not figure in)
$info{cursx}=0;			# x-position of the cursor (insertion point)
$info{cursy}=0;			# y-position of the cursor
$info{position}=0;		# line num in buffer on the top of the screen

initscr;
cbreak;				# disable buffering and erase/kill
# once we're stable enough to ignore ^C we should probably go raw
noecho;				# disable getch echoing
timeout(-1);			# getch blocks for input
keypad(stdscr,1);		# decode extended keys to curses constants

# $LINES doesn't get a value until after we initscr
$info{winlines}=$LINES-2;

&{$$cinfo{statfunc}};
$diedie=0;

if(-e '/etc/acedrc') {
    do('/etc/acedrc');
}
if(-e "$ENV{HOME}/.acedrc") {
    do("$ENV{HOME}/.acedrc");
}

$startline=0;
while(defined($_=shift)) {
#    &statmsg("Argument: $_");
#    sleep(2);
    if($_=~/^\+(\d+)/) {
	$startline=$1;
    } else {
	$startopen=$_;
    }
}
if(defined($startopen)) {
    &load($startopen);
    &zaptoline($startline);
}

while($diedie==0) {
    $inchar=getch;
    if($inchar eq $lastchar) {
	$repeated=1;
    } else {
	$repeated=0;
    }
    if(defined($$cinfo{bindings}{$inchar})) { # key is bound to a function
	&{$$cinfo{bindings}{$inchar}};
    } else {			# no binding, just insert this character
	&insatcurs($inchar);
#	&statmsg("would've inserted $inchar");
	refresh;
    }
    $lastchar=$inchar;
}
endwin;

# draw the status bar
sub statbar {
    addstr($$cinfo{winlines}+$$cinfo{wintop},0,"--aced--");
#    addstr($LINES-2,0,"--aced--");
#    clrtobot;
    clrtoeol;
    &movecurs();
}

# draw a message under the status line
sub statmsg {
    addstr($LINES-1,0,$_[0]);
    clrtobot;
    &movecurs();
}

# display an error message - statmsg but with a beep
sub errmsg {
    beep;
    &statmsg($_[0]);
}

# get a line of user input outside the edit region
# optional argument: prompt
sub getline {
    my $line;
    my $y;

    echo;
    move($LINES-1,0);
    if(defined($_[0])) {
	addstr($_[0]);
    }
    clrtobot;
    getstr($line);
    noecho;
    
    return $line;
}

# determine what line of the buffer we're on
sub bufline {
    return $$cinfo{cursy}+$$cinfo{position};
}

# insert character(s) at the cursor and advance the cursor
sub insatcurs {
    substr($$cbuf[&bufline],$$cinfo{cursx})=
	$_[0].substr($$cbuf[&bufline],$$cinfo{cursx});
    $$cinfo{cursx}+=length($_[0]);
    &refreshbuf(); # fixme only needs to refresh this line
}

# refresh the image of the buffer on the screen
sub refreshbuf {
    my $curstr;

    for($i=0;$i<$$cinfo{winlines};$i++) {
	$curstr=&escchars($$cbuf[$i+$$cinfo{position}]);
	addstr($i+$$cinfo{wintop},0,$curstr);
	clrtoeol;
    }
    &movecurs();
}

# move the cursor 1 character forward
sub cursfwd {
    if($$cinfo{cursx}==length($$cbuf[&bufline])) {
	if(defined($$cbuf[&bufline+1])) {
	    $$cinfo{cursx}=0;
	    &cursdown();
	} else {
	    &errmsg("End of buffer");
	}
    } else {
	$$cinfo{cursx}++;
	&movecurs();
    }
}

# move the cursor 1 character back
sub cursback {
    if($$cinfo{cursx}==0) {
	if($$cinfo{cursy}>0) {
	    &cursup();
	    $$cinfo{cursx}=length($$cbuf[&bufline]);
	} else {
	    &errmsg("Beginning of buffer");
	}
    } else {
	$$cinfo{cursx}--;
    }
    &movecurs();
}

# move the cursor to the beginning of the current line
sub cursbol {
    $$cinfo{cursx}=0;
    &movecurs();
}

# move the cursor to the end of the current line
sub curseol {
    $$cinfo{cursx}=length($$cbuf[&bufline]);
    &movecurs();
}

# split the current line at the cursor
sub insnewline {
    splice(@$cbuf,&bufline+1,0,substr($$cbuf[&bufline],$$cinfo{cursx}));
    substr($$cbuf[&bufline],$$cinfo{cursx})='';
    &refreshbuf;
    &cursfwd;
}

# insert string(s) at the cursor, interpreting newlines properly
# multiple arguments are concatenated so you can do &paste(`ls`)
sub paste {
    my $char;

    foreach $char (split(//,join('',@_))) {
	if($char=~/\r|\n/) {
	    &insnewline();
	} else {
	    &insatcurs($char);
	}
    }
}

# delete the character to the left of the cursor
sub backspace {
    if($$cinfo{cursx}==0) {
	if(&bufline==0) {
	    &errmsg("Beginning of buffer");
	} else {
	    $$cinfo{cursx}=length($$cbuf[&bufline-1]);
	    splice(@$cbuf,&bufline-1,2,$$cbuf[&bufline-1].$$cbuf[&bufline]);
	    $$cinfo{cursy}--;
	}
    } else {
	substr($$cbuf[&bufline],$$cinfo{cursx}-1)=
	    substr($$cbuf[&bufline],$$cinfo{cursx});
	$$cinfo{cursx}--;
    }
    &refreshbuf();
}

# delete the character under the cursor
sub delunder {
    if($$cinfo{cursx}==length($$cbuf[&bufline])) {
	if(defined($$cbuf[&bufline+1])) {
	    splice(@$cbuf,&bufline,2,$$cbuf[&bufline].$$cbuf[&bufline+1]);
	} else {
	    &errmsg("End of buffer");
	}
    } else {
	substr($$cbuf[&bufline],0,$$cinfo{cursx}+1)=
	    substr($$cbuf[&bufline],0,$$cinfo{cursx});
    }
    &refreshbuf();
}

# Move the cursor down one line
sub cursdown {
    if(defined($$cbuf[&bufline+1])) {
	&zaptoline(&bufline+1);
    } else {
	&errmsg("End of buffer");
    }
}

# Move the cursor up one line
sub cursup {
    if(&bufline==0) {
	&errmsg("Beginning of buffer");
    } else {
	&zaptoline(&bufline-1);
    }
}

# Save the current buffer. If an argument is given, use that as the filename,
# else prompt the user. If an argument is given, prompt before overwriting.
# FIXME we should avoid race conditions when writing to /tmp and such
sub saveas {
    my($filename, $sure, $line);

    if(defined($_[0])) {
	$filename=$_[0];
    } else {
	$filename=&getline("Filename: ");
    }

    if(-e $filename) {
	$sure=&getline("Sure you want to overwrite $filename? ");
	if($sure!~/^\s*[Yy]/) {
	    &statmsg("Cancelled");
	    return;
	}
    }

    open(FOO,">$filename");
    foreach $line (@$cbuf) {
	print FOO "$line\n";
    }
    close FOO;
}

# Execute a subcommand, either by suspending curses or in an xterm.
sub shell {
    my $command;

    if($_[0]) {
	$command=$_[0];
    } else {
	$command="$ENV{SHELL} -i";
    }

    if($ENV{DISPLAY}) {
	system("xterm -e $ENV{SHELL} -c '$command' &");
    } else {
	endwin;
	system($command);
	refresh;
    }
}

# Scroll the buffer up a page
sub curspgup {
    $$cinfo{cursy}=0;
    &cursup();
}

# Scroll the buffer down a page
sub curspgdn {
    $$cinfo{cursy}=$$cinfo{winlines}-1;
    if(&bufline>$#$cbuf) {
	$$cinfo{cursy}=0; # FIXME
    }
    &cursdown();
}

# Erase the current line and put it in the cut buffer
sub cutline {
    if(!$repeated) {
	$cutbuffer='';
    }

    if($$cbuf[&bufline]=~/^$/) {
	&delunder();
	$cutbuffer.="\n";
    } else {
	$cutbuffer.=$$cbuf[&bufline];
	$$cbuf[&bufline]='';
	$$cinfo{cursx}=0;
	&refreshbuf();
    }
}

# indent to a multiple of 5 characters
sub tab {
baz:for(;;) {
	&paste(" ");
	($$cinfo{cursx}=~/[05]$/) and last baz;
    }
}

# load a file. if an argument is given, use it. else prompt.
sub load {
    my $filename;

    if(defined($_[0])) {
        $filename=$_[0];
    } else {
        $filename=&getline("Filename: ");
    }

    if(!-e $filename) {
	&errmsg("No such file");
    } else {
	open(FOO,"<$filename");
# unfortunately we need to drop the \ns ourselves instead of just @$cbuf=<FOO>
	@$cbuf=();
	while(<FOO>) {
	    chomp;
	    push @$cbuf,$_;
	}
	close FOO;
	$$cinfo{cursx}=0;
	$$cinfo{cursy}=0;
	$$cinfo{position}=0;
	&refreshbuf();
    }
}

# find a regex in the current buffer
sub find {
    my $regex;
    my $i;
    my $curend;

    if(defined($_[0])) {
        $regex=$_[0];
    } else {
        $regex=&getline("Regex: ");
    }

    # check current line after cursor
    if($$cinfo{cursx}<length($$cbuf[&bufline])) {
	$curend=substr($$cbuf[&bufline],$$cinfo{cursx}+1);
	if($curend=~/$regex/) {
	    &findfound(&bufline,index($$cbuf[&bufline],$&,$$cinfo{cursx}+1));
#	    &statmsg("Found on current line");
	    return;
	}
    }

    # check file after this line
    for($i=&bufline+1;defined($$cbuf[$i]);$i++) {
	if($$cbuf[$i]=~/$regex/) {
	    &findfound($i,index($$cbuf[$i],$&));
#	    &statmsg("Found after current line");
	    return;
	}
    }

    # check file before this line
    # also check this line (we ignored the beginning part before)
    for($i=0;$i<=&bufline;$i++) {
	if($$cbuf[$i]=~/$regex/) {
	    &statmsg("Search wrapped");
	    &findfound($i,index($$cbuf[$i],$&));
#	    &statmsg("Found before current line");
	    return;
	}
    }

    &statmsg("Not found")
}

# called when &find() finds a string: &findfound(line,position)
sub findfound {
#    &statmsg("found at $_[0] $_[1]");
    &zaptoline($_[0]);
    $$cinfo{cursx}=$_[1];
    &refreshbuf();
}

# move the cursor to the specified line, scrolling the buffer if needed
sub zaptoline {
    if($_[0]<$$cinfo{position} || $_[0]>=$$cinfo{position}+$$cinfo{winlines}) {
	$$cinfo{position}=$_[0];
    }

    $$cinfo{cursy}=$_[0]-$$cinfo{position};

    if($$cinfo{cursx}>length($$cbuf[&bufline])) {
	$$cinfo{cursx}=length($$cbuf[&bufline]);
	&refreshbuf();
    }

    &refreshbuf();
}

# Move the cursor to the top of the file
sub cursbof {
    $$cinfo{cursx}=0;
    &zaptoline(0);
}

# Move the cursor to the eof
sub curseof {
    &zaptoline($#$cbuf);
    $$cinfo{cursx}=length($$cbuf[&bufline]);
}

# Given a string, escape non-printables
sub escchars {
    my $foo=$_[0];

    $foo=~s/([\x00-\x1F])/&esc1char($1)/ge;
    return $foo;
}

# Convert one control character to its printable equivalent
sub esc1char {
    return '\x'.sprintf("%02X",ord($_[0]));
}

# Put the cursor in the right place
sub movecurs {
    move($$cinfo{cursy}+$$cinfo{wintop},
	 length(&escchars(substr($$cbuf[&bufline],0,$$cinfo{cursx}))));
    refresh;
}
